#!/usr/bin/env python3

#
# Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of NVIDIA CORPORATION &
# AFFILIATES (the "Company") and all right, title, and interest in and to the
# software product, including all associated intellectual property rights, are
# and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
#

import subprocess
import json
import sys
import struct
import hashlib
import itertools
import os
import os.path
import shutil
import argparse
import psutil, shutil, stat
from ctypes import *
import importlib
import binascii
import datetime
import json
import logging
import os
from typing import Dict, Union, Optional, Any, Set
from urllib import request
from pathlib import Path
import pdbparse
import pdbparse.undecorate
import http.client as httplib
import pefile
import binascii
from pefile import PE

class LinuxApsh():

    def shell_cmd(self, cmd):
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        (out, err) = map(lambda x: (x or b'').decode('utf-8'), proc.communicate())
        return (out, err)

    # hash_build run
    def run_hash_build(self, args):

        if os.path.exists("hash.zip"):
            os.remove("hash.zip")

        if os.path.exists("apsh_client_build_hash"):
            shutil.rmtree("apsh_client_build_hash")

        os.mkdir("apsh_client_build_hash")

        with open("/proc/" + str(args.pid) + "/maps", 'r') as map_file:
            files = list(set([line.strip().split()[-1] for line in map_file.readlines() if len(line.strip().split()) == 6]))
            files = [file for file in files if not file.startswith('[')]
            for file in files:
                shutil.copy(file, "apsh_client_build_hash")

        shutil.make_archive("hash", 'zip', "apsh_client_build_hash")

    #   Symbols
    def create_symbols(self, args):
        # get linux distribution
        (out, err) = self.shell_cmd("cat /etc/os-release | grep ^ID=")
        dist = out.strip().split("=")[1].strip("\"")
        print("OS distribution = " + dist) 

        # get kernel release
        kernel_release = os.uname()[2]

        if not os.path.isfile(args.path):
            print("Dwarf2Json not found. Please specify path to Dwarf2Json executable using --path")
            print("For download and compilation guide, please visit https://github.com/volatilityfoundation/dwarf2json")
            exit(-1)

        if dist.lower() == 'ubuntu':
            print("Recognized UbuntuOS - Starting....")
            ubuntu_vmlinux_file="/usr/lib/debug/boot/vmlinux-" + kernel_release
            ubuntu_system_map_file="/boot/System.map-" + kernel_release
            if os.path.isfile(ubuntu_vmlinux_file) and os.path.isfile(ubuntu_system_map_file):
                print("found current kernel " + kernel_release + " debug files...")
                print("executing dwarf2json...")
                cmd = "sudo " + args.path + " linux --elf " +ubuntu_vmlinux_file+ " --system-map "+ubuntu_system_map_file+" > ./symbols.json"
                (out, err) = self.shell_cmd(cmd)
            else:
                print("identified ubuntu system, installing debug symbols...")
                lsb_release = self.shell_cmd("lsb_release -cs")
                with open("/etc/apt/sources.list.d/ddebs.list", 'a') as file:
                    file.write("deb http://ddebs.ubuntu.com $(lsb_release -cs) main restricted universe multiverse\n" +
                                "deb http://ddebs.ubuntu.com $(lsb_release -cs)-updates main restricted universe multiverse\n" +
                                "deb http://ddebs.ubuntu.com $(lsb_release -cs)-proposed main restricted universe multiverse\n")
                self.shell_cmd("sudo apt update && apt -y install ubuntu-dbgsym-keyring linux-image-"+kernel_release+"-dbgsym")
                if os.path.isfile(ubuntu_vmlinux_file) and os.path.isfile(ubuntu_system_map_file):
                    print("executing dwarf2json...")
                    cmd = "sudo " + args.path + " linux --elf " +ubuntu_vmlinux_file+ " --system-map "+ubuntu_system_map_file+" > ./symbols.json"
                    (out, err) = self.shell_cmd(cmd)
                else:
                    print("vmlinux and system map file not found")

        if dist.lower() == 'centos':
            print("Recognized CentOS - Starting....")
            centos_vmlinux_file="/usr/lib/debug/lib/modules/"+kernel_release+"/vmlinux"
            centos_system_map_file="/boot/System.map-"+ kernel_release
            if os.path.isfile(centos_vmlinux_file) and os.path.isfile(centos_system_map_file):
                print("found current kernel " + kernel_release + " debug files...")
                print("executing dwarf2json...")
                cmd = "sudo " + args.path + " linux --elf " +centos_vmlinux_file+ " --system-map "+centos_system_map_file+" > ./symbols.json"
                (out, err) = self.shell_cmd(cmd)
            else:
                print("identified Centos system, installing debug symbols...")
                self.shell_cmd("sed -i 's/enabled=0/enabled=1/g' /etc/yum.repos.d/CentOS-Linux-Debuginfo.repo && yum -y install kernel-debuginfo")
                if os.path.isfile(centos_vmlinux_file) and os.path.isfile(centos_system_map_file):
                    print("executing dwarf2json...")
                    cmd = "sudo " + args.path + " linux --elf " +centos_vmlinux_file+ " --system-map "+centos_system_map_file+" > ./symbols.json"
                    (out, err) = self.shell_cmd(cmd)
                else:
                    print("vmlinux and system map file not found")

        print("cleaning and exit.")


    def create_mem_regions(self, args):
        (out, err) = self.shell_cmd("sudo less /proc/iomem | grep \"System RAM\"")
        dict1 = {}
        output_string = "{\n   \"allowed_regions\" : {"

        for i, line in enumerate(out.splitlines()):
            addresses = line.strip().split()[0]
            start = int(addresses.split('-')[0], 16)
            end = int(addresses.split('-')[1], 16)
            line =  "\n      \""+str(i)+"\" : {" \
                    +"\n         \"length\" : "+str(end - start)+"," \
                    +"\n         \"start\" : "+str(start) \
                    +"\n      },"
            output_string += line

        output_string += "\n   }\n}\n"
        if os.path.exists("mem_regions.json"):
            os.remove("mem_regions.json")
        with open("mem_regions.json", 'w') as file:
            file.write(output_string)


    def __init__(self, args):
        for func in args.files:
            getattr(self, func_map[func])(args)

class WindowsApsh():
    # hash_build run
    def run_hash_build(self, args):

        if os.path.exists("hash.zip"):
            os.remove("hash.zip")

        if os.path.exists("apsh_client_build_hash"):
            self.rmtree("apsh_client_build_hash")

        os.mkdir("apsh_client_build_hash")

        for map_file in psutil.Process(args.pid).memory_maps():
            shutil.copy(map_file.path, "apsh_client_build_hash")

        shutil.make_archive("hash", 'zip', "apsh_client_build_hash")

    MAX_MEM_REGIONS = 20

    CM_RESOURCE_MEMORY_LARGE = 0x0E00
    CM_RESOURCE_MEMORY_LARGE_40 = 0x0200
    CM_RESOURCE_MEMORY_LARGE_48 = 0x0400
    CM_RESOURCE_MEMORY_LARGE_64 =  0x0800

    CmResourceTypeNull = 0 # ResType_All or ResType_None (0x0000)
    CmResourceTypePort = 1 # ResType_IO (0x0002)
    CmResourceTypeInterrupt = 2 # ResType_IRQ (0x0004)
    CmResourceTypeMemory = 3 # ResType_Mem (0x0001)
    CmResourceTypeDma = 4 # ResType_DMA (0x0003)
    CmResourceTypeDeviceSpecific = 5 # ResType_ClassSpecific (0xFFFF)
    CmResourceTypeBusNumber = 6 # ResType_BusNumber (0x0006)
    CmResourceTypeMemoryLarge = 7 # ResType_MemLarge (0x0007)
    CmResourceTypeNonArbitrated = 128 # Not arbitrated if 0x80 bit set
    CmResourceTypeConfigData = 128 # ResType_Reserved (0x8000)
    CmResourceTypeDevicePrivate = 129 # ResType_DevicePrivate (0x8001)
    CmResourceTypePcCardConfig = 130 # ResType_PcCardConfig (0x8002)
    CmResourceTypeMfCardConfig = 131 # ResType_MfCardConfig (0x8003)
    CmResourceTypeConnection = 132 # ResType_Connection (0x8004)

    def parse_memory_map():
        winreg = importlib.import_module('winreg')
        pszSubKey = "Hardware\\ResourceMap\\System Resources\\Physical Memory"
        pszValueName = ".Translated"
        key = winreg.OpenKeyEx(winreg.HKEY_LOCAL_MACHINE, pszSubKey)
        dw = winreg.QueryValueEx(key, pszValueName)
        regions = []
        for x in range(20, len(dw[0]), 20):
            start = int.from_bytes(dw[0][x+4:x+12], byteorder='little')
            length = int.from_bytes(dw[0][x+12:x+20], byteorder='little')
            flag = int.from_bytes(dw[0][x+2:x+4], byteorder='big')
            if flag==2: length*=256
            if flag==4: length*=256**2
            if flag==8: length*=256**4
            regions.append({
                "start": start,
                "length": length
            })
        winreg.CloseKey(key)
        return regions


    def create_mem_regions(self, args):
        regions = WindowsApsh.parse_memory_map()
        json = {"allowed_regions": {}}
        for i,r in enumerate(regions):
            json["allowed_regions"][str(i)] = r
        if os.path.exists("mem_regions.json"):
            os.remove("mem_regions.json")
        with open("mem_regions.json", 'w') as file:
            file.write(str(json))

    def get_guid():
        pe = PE("C:\\Windows\\System32\\ntoskrnl.exe")
        debug = pe.DIRECTORY_ENTRY_DEBUG[0].entry
        guid = "{0:08X}{1:04X}{2:04X}{3}{4}".format(debug.Signature_Data1,
        debug.Signature_Data2,
        debug.Signature_Data3,
        binascii.hexlify(debug.Signature_Data4).decode("utf-8"),
        debug.Age).lower()
        return guid

    logger = logging.getLogger(__name__)
    logger.setLevel(1)

    def load_pdbparse(self, args):
        if os.path.isfile(args.path):
            file_directory, file_name = os.path.split(args.path)
            sys.path.append(file_directory)
            return importlib.import_module(file_name.split('.')[0])
        else:
            do_download = False
            valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
            question = 'pdbparse-to-json.py not found.\ndo you want to download?'
            prompt = " [y/N] "
            default = 'n'
            while True:
                sys.stdout.write(question + prompt)
                choice = input().lower()
                if choice == "":
                    do_download = valid[default]
                    break
                elif choice in valid:
                    do_download = valid[choice]
                    break
                else:
                    sys.stdout.write("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")
            if do_download:
                import urllib.request
                pdbparse_url = 'https://raw.githubusercontent.com/volatilityfoundation/volatility3/stable/development/pdbparse-to-json.py'
                urllib.request.urlretrieve(pdbparse_url, "pdbparse-to-json.py")
                sys.path.append(os. getcwd())
                return importlib.import_module('pdbparse-to-json')

    def create_symbols(self, args):
        pdbparse = self.load_pdbparse(args)
        filename = pdbparse.PDBRetreiver().retreive_pdb(WindowsApsh.get_guid(), "ntkrnlmp.pdb")

        delfile = True
        if not filename:
            parser.error("No suitable filename provided or retrieved")
        convertor = pdbparse.PDBConvertor(filename)
        newfilename = "symbols.json"
        p = Path(newfilename)
        Path(p.parent).mkdir(parents=True, exist_ok=True)
        with open(newfilename, "w") as f:
            json.dump(convertor.read_pdb(), f, indent = 2, sort_keys = True)
        print("Temporary PDB file: {}".format(newfilename))

    def __init__(self, args):
        for func in args.files:
            getattr(self, func_map[func])(args)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='', formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("--os", type=str, help='os to run on - windows or linux')
    func_map = {'hash': "run_hash_build",
            'symbols': "create_symbols",
            'memregions': "create_mem_regions"
            }
    functions = []
    default_dwarf2json_path = './dwarf2json'
    default_pdbparse_path = './pdbparse-to-json.py'

    parser.add_argument("pid", type=int, help='an integer for the accumulator')
    parser.add_argument('--files',nargs='+',metavar='N',choices=list(func_map.keys()), default=list(func_map.keys()))
    parser.add_argument('--path', help="path to Dwarf2Json executble (for linux os) or pdbparse-to-json.py (for windows).\n\n" +
        "Dwarf2Json - download and compilation guide can be found at:\nhttps://github.com/volatilityfoundation/dwarf2json\n\n" +
        "pdbparse-to-json.py - can be found at:\nhttps://raw.githubusercontent.com/volatilityfoundation/volatility3/stable/development/pdbparse-to-json.py")
    args = parser.parse_args()
    print('creating:' + str(args.files))
    args = parser.parse_args()
    if args.os == "linux":
        if not args.path:
            args.path = default_dwarf2json_path
        LinuxApsh(args)
    elif args.os == "windows":
        if not args.path:
            args.path = default_pdbparse_path
        WindowsApsh(args)
    else:
        print("unsupported os")
